/* elflib.c -- analyze ELF executable or shared library
   Copyright 2000 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/

#include <elf.h>
#include <stdio.h>  /* perror */
#include <stdlib.h>  /* free */
#include <string.h>  /* strncmp */

#include "elflib.h"  /* prototypes for self */
#include "syslib.h"  /* getfile */

int
elf_good_header(Elf32_Ehdr const *const ehdr)
{
	if (0!=strncmp((void *)ehdr, ELFMAG, SELFMAG)
	||  ELFCLASS32  != ehdr->e_ident[EI_CLASS]
	||  ELFDATA2LSB != ehdr->e_ident[EI_DATA]
	||  EV_CURRENT != ehdr->e_ident[EI_VERSION]
	|| (EM_386 != ehdr->e_machine /*&& EM_486 != ehdr->e_machine*/)
	|| (ET_EXEC != ehdr->e_type && ET_DYN != ehdr->e_type) ) {
		return 0;
	}
	return 1;
}

Elf32_Ehdr *
elf_mmap(
	char const *const elf_name,
	size_t *p_elf_size
)
{
	Elf32_Ehdr *const ehdr = (Elf32_Ehdr *)getfile(elf_name, p_elf_size, 1);
	if (0==elf_good_header(ehdr)) {
		perror(elf_name);
		exit(1);
	}
	return ehdr;
}

Elf32_Phdr const *
elf_find_phdr_type(
	int type,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Phdr const *phdr = (Elf32_Phdr const *)(ehdr->e_phoff + (char *)ehdr);
	int j = ehdr->e_phnum;
	for (; --j>=0; ++phdr) {
		if (type==phdr->p_type) {
			return phdr;
		}
	}
	return 0;
}


Elf32_Shdr const * elf_find_section(
	void const *const ptr,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Shdr const *shdr = (Elf32_Shdr const *)(ehdr->e_shoff + (char *)ehdr);
	int j = ehdr->e_shnum;
	for (; --j>=0; ++shdr) {
		if (ptr==(shdr->sh_offset + (char *)ehdr)) {
			return shdr;
		}
	}
	return 0;
}

Elf32_Shdr const *
elf_find_section_name(
	char const *const name,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Shdr const *shdr = (Elf32_Shdr const *)(ehdr->e_shoff + (char *)ehdr);
	Elf32_Shdr const *const shstr = ehdr->e_shstrndx + shdr;
	int j = ehdr->e_shnum;
	for (; --j>=0; ++shdr) {
		char const *const sh_name = shdr->sh_name + shstr->sh_offset + (char *)ehdr;
		if (0==strcmp(name, sh_name)) {
			return shdr;
		}
	}
	return 0;
}

Elf32_Shdr const *
elf_find_section_type(
	int key,
	Elf32_Ehdr const *const ehdr
)
{
	int j = ehdr->e_shnum;
	Elf32_Shdr const *shdr = (Elf32_Shdr const *)(ehdr->e_shoff + (char *)ehdr);
	for (; --j>=0; ++shdr) {
		if (shdr->sh_type == key) {
			return shdr;
		}
	}
	return 0;
}

unsigned
elf_get_dynamic(
	int const key,
	Elf32_Dyn const *dynp
)
{
	for (; DT_NULL!=dynp->d_tag; ++dynp) {
		if (dynp->d_tag == key) {
			return dynp->d_un.d_val;
		}
	}
	return ~0;
}

unsigned
elf_set_dynamic(
	int const key,
	unsigned const value,
	Elf32_Dyn *dynp
)
{
	for (; DT_NULL!=dynp->d_tag; ++dynp) {
		if (dynp->d_tag == key) {
			dynp->d_un.d_val = value;
			return 0;
		}
	}
	return ~0;
}

Elf32_Dyn const *
elf_locate_dynamic(
	int const key,
	Elf32_Dyn const *dynp
)
{
	for (; DT_NULL!=dynp->d_tag; ++dynp) {
		if (dynp->d_tag == key) {
			return dynp;
		}
	}
	return 0;
}

void const *
elf_find_dynamic(
	int const key,
	Elf32_Dyn const *dynp,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Phdr const *const pt_text = elf_find_phdr_type(PT_LOAD, ehdr);
	unsigned const tx_reloc = pt_text->p_vaddr - pt_text->p_offset;
	for (; DT_NULL!=dynp->d_tag; ++dynp) {
		if (dynp->d_tag == key) {
			return (void const *)(dynp->d_un.d_val - tx_reloc +
				(char *)ehdr );
		}
	}
	return 0;
}

unsigned
elf_get_offset(
	unsigned const offset,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Phdr const *phdr = (Elf32_Phdr const *)(ehdr->e_phoff + (char *)ehdr);
	int j = ehdr->e_phnum;
	for (; --j>=0; ++phdr) if (PT_LOAD == phdr->p_type) {
		unsigned const t = offset - phdr->p_vaddr;
		if (t < phdr->p_filesz) {
			return t + phdr->p_offset;
		}
	}
	return 0;
}

char *
elf_find_offset(
	unsigned const offset,
	Elf32_Ehdr const *const ehdr
)
{
	unsigned const t = elf_get_offset(offset, ehdr);
	if (t) {
		return t + (char *)ehdr;
	}
	return 0;
}

unsigned
elf_find_value(
	Elf32_Sym const *const sp,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Shdr const *shdr = (Elf32_Shdr const *)(ehdr->e_shoff + (char *)ehdr);
	if (STB_LOCAL == ELF32_ST_BIND(sp->st_info))
	if (sp->st_shndx < ehdr->e_shnum) {
		return sp->st_value + elf_get_offset(shdr[sp->st_shndx].sh_addr, ehdr);
	}
	return sp->st_value;
}

unsigned int
elf_hash_dynsym(char const *p)
{
	unsigned int h;
	for (h = 0; 0 != *p; ++p) {
		h = *p + (h<<4);
	    {
		unsigned int const t = 0xf0000000U & h;
		if (0!=t) {/* test could be omitted! */
			h &= ~t;
			h ^= t >> 24;
		}
	    }
	}
	return h;
}

void
elf_insert_dynsym(
	char const *const name,
	unsigned const ndx,
	unsigned *const htab
)
{
	unsigned int const nbucket = htab[0];
	/*unsigned int const nchain  = htab[1];*/
	unsigned int *const buckets = &htab[2];
	unsigned int *const chains  = &buckets[nbucket];
	unsigned int const m = elf_hash_dynsym(name) % nbucket;

	if (0!=buckets[m]) {
		chains[ndx] = buckets[m];
	}
	buckets[m] = ndx;
}

void
elf_rehash(
	unsigned       *const new_htab,
	unsigned const *const htab,
	Elf32_Sym const *const dynsym,
	char const *const dynstr
)
{
	unsigned const nbucket = htab[0];
	unsigned const nchain  = htab[1];
	unsigned const *const bucket = &htab[2];
	unsigned const *const chain  = &bucket[nbucket];
	unsigned *const tempch = (unsigned *)zalloc(sizeof(unsigned) * nchain);
	int j;
	for (j=nbucket; --j>=0; ) { /* transfer from old hash table to new */
		unsigned *t = tempch;
		int k;
		*t++ = 0;  /* sentinel */
		for (k=bucket[j]; 0!=k; k=chain[k]) { /* read chain */
			*t++ = k;
		}
		while (0!=(k = *--t)) { /* write, keeping original order */
			elf_insert_dynsym(dynsym[k].st_name+dynstr, k, new_htab);
		}
	}
	free(tempch);
}

