/* expvec.c -- build an export transfer vector for an ELF shared library
   Copyright 2000,2001 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <sys/types.h>
#include <sys/stat.h>  /* umask */
#include <utime.h>  /* utime */

#include <elf.h>
#include <regex.h>  /* regexec */
#include <stdio.h>  /* FILENAME_MAX fprintf rename stderr */
#include <stdlib.h>  /* mkstemp */
#include <string.h>  /* strcat */
#include <unistd.h>  /* unlink */

#include "elflib.h"
#include "syslib.h"
#include "veclib.h"

static char *
do_asm(char *const tmpname)
{
	static char o_name[FILENAME_MAX];
	       char s_name[FILENAME_MAX];
	char const *argv[6];
	argv[0] = "cc";
	argv[1] = "-c";
	argv[2] = "-o";
	argv[3] = strcat(strcpy(o_name, tmpname), ".o");
	argv[4] = strcat(strcpy(s_name, tmpname), ".s");
	argv[5] = 0;

	rename(tmpname, s_name);
    {
	mode_t const oldmask = umask(077);  /* paranoia */
	do_exec(argv, 0,0,0);
	umask(oldmask);
    }
	return o_name;
}

static void
bad_sync(char const *const prog)
{
	fprintf(stderr,"%s: warning; could not sync time from fd 3 to fd 1\n",
		prog );
}

static int
imax(int a, int b)
{
	if (a <= b) {
		return b;
	}
	return a;
}

static int
detect_rebuild(int const argc, char const *const arg1, char ***p_argv)
{
	if (0==strncmp("--rebuild", arg1, imax(5, strlen(arg1)))) {
		(*p_argv)[1] = (*p_argv)[0];
		++*p_argv;
		return 1;
	}
	return 0;
}

int
main(int argc, char **argv)
{
	/*  expvec [--rebuild] sym_order non_funcs version_script loader_args...  \
	    > import_order  3> compiled_export_vector
	*/

	int const rebuild = detect_rebuild(argc, argv[1], &argv);

	/* first pass: construct .so from loader_args */
	int const junk = do_exec((char const *const *)(4+argv), 0,0,0);

	/* find output file */
	char const *const *const plibfn = find_outfn((char const *const *)(4+argv));

	/* get prefered symbol order, and the excluded symbols */
	regex_t const */*const*/ sympats = get_symbol_patterns(argv[1], rebuild);
	regex_t const */*const*/ exclude = get_symbol_patterns(argv[2], 0);

	/* map the .so, and locate interesting pieces */
	size_t libsz = 0;
	Elf32_Ehdr const *const ehdr = elf_mmap(*plibfn, &libsz);
	Elf32_Shdr const *const dynsec = elf_find_section_type(SHT_DYNAMIC, ehdr);
	Elf32_Dyn const *const dynamic = (Elf32_Dyn const *)(dynsec->sh_offset + (int)ehdr);

	Elf32_Shdr const *const symsec = elf_find_section_type(SHT_DYNSYM, ehdr);
	Elf32_Sym const *const dynsym = (Elf32_Sym const *)elf_find_dynamic(DT_SYMTAB, dynamic, ehdr);
	int const n_dynsym = symsec->sh_size / sizeof(Elf32_Sym);
	int       r_dynsym = n_dynsym;  /* number of symbols remaining */
	int *d_dynsym = (int *)zalloc(sizeof(int) * n_dynsym); /* status of each symbol */
		/* -1: excluded;  else: position in vector */

	char const *const strtab = (char const *)elf_find_dynamic(DT_STRTAB, dynamic, ehdr);

	char const *const base = "___base";
	char const *const vecname = get_vecname(strtab, dynamic, *plibfn);
	char *const tmpname = my_mktmpname("vl_");
	FILE *const vecf = fdopen(mkstemp(tmpname), "w");
	int n_symexcl;  /* number of symbols actually excluded */
	static regex_t onepat[2];  /* onepat[1] is always 0 */
	regex_t const *pat;
	int j;

	(void)junk;
	printf("%s\n",vecname);

	/* exclude non-functions and user-excluded functions */
	for (j=0; j<n_dynsym; ++j) {
		Elf32_Sym const *const symp = j + dynsym;
		int const type = ELF32_ST_TYPE(symp->st_info);
		if (SHN_UNDEF==symp->st_shndx || SHN_ABS==symp->st_shndx
		|| (STT_FUNC!=type && STT_NOTYPE!=type)
		|| 0==symp->st_name ) {
			d_dynsym[j] = -1;
		}
		else {
			char const *const name = symp->st_name + strtab;
			d_dynsym[j] = 0 - matches(name, exclude);
		}
		r_dynsym += d_dynsym[j];  /* decrement, or same */
	}
	n_symexcl = n_dynsym - r_dynsym;

    if (0!=r_dynsym) {
	fprintf(vecf,
		"	.section \".rodata\"; .align 1<<2\n"
		"	movl $0,%%ebx\n"
		"%s: .globl %s\n\t.type %s,@function; .size %s,11\n",
			vecname, vecname, vecname, vecname);
	fprintf(vecf,
		"	call __reloc\n"
		"	ret\n"
		"	nop\n"
		"__reloc:\n"
		"	addl (%%esp),%%ebx\n"
		"	ret\n"
		"%s:\n",
			base );
    }
    {
	/* order the functions */
	int n_sympat = 0;
	char const **patlit;  /* rebuild mode only */
	if (rebuild) {
		char const *sym0 = *(char const **)sympats;
		if (0!=strncmp(sym0, "__V_", 4)) {
			fprintf(stderr, "warning: __V_ vector name is not the first symbol in --rebuild mode\n");
		}
	}
	if (0!=r_dynsym)
	for (
	  (pat = sympats), (patlit = (char const **)sympats);
	  0!=(rebuild ? *patlit : (char const *)pat->buffer);
	  ++pat, ++patlit
	) {
		++n_sympat;
		if (0==rebuild) {
			onepat[0] = *pat;
		}
		for (j=0; j<n_dynsym; ++j) if (0==d_dynsym[j]) {
			char const *const name = dynsym[j].st_name + strtab;
			if ((0==rebuild && matches(name, &onepat[0]))
			||  (0!=rebuild && 0==strcmp(name, *patlit)) ) {
				fprintf(vecf,".long %s - %s\n",
					name, base);
				printf("%s\n", name);
				d_dynsym[j] = r_dynsym--;
				if (rebuild) { /* name can be used only once */
					char const *const name = *patlit;
					*patlit = strlen(name) + name;
				}
				if (0==r_dynsym) {
					fprintf(vecf,".long 0\n");
					goto ordered;
				}
			}
		}
	}
ordered: ;
	if (rebuild) { /* check that each name was found */
		char const *const *patlit = (char const *const *)sympats;
		int n_fatal = 0;
		for (j=n_sympat; --j>=0; ++patlit) {
			char const *const name = *patlit;
			if (0!=name[0]
			&& (  (1+ j)!=n_sympat
			   || 0!=strncmp(name, "__V_", 4) ) ) {
				++n_fatal;
				fprintf(stderr,
					"fatal: rebuild did not find %s\n",
					name );
			}
		}
		if (0!=n_fatal) {
			exit(1);
		}
	}
	if (0!=r_dynsym) {
		fprintf(stderr,"note: %d unordered functions\n", r_dynsym);
		for (j=0; j<n_dynsym; ++j) if (0==d_dynsym[j]) {
			char const *const name = dynsym[j].st_name + strtab;
			fprintf(stderr, "\t%s\n", name);
			fprintf(vecf,".long %s - %s\n",
				name, base);
			printf("%s\n", name);
		}
		fprintf(vecf,".long 0\n");
	}
	fclose(vecf);
	fflush(stdout);  /* to avoid write after sync with fd 3 */
    }
	
    {
	/* Splice the export vector .o into the commandline just ahead of the "-o". */
	char *const objname = do_asm(tmpname);
	char *const cmd = argv[4];
	char new_vscript[FILENAME_MAX];
	char const *      *q = (char const *      *)(2+argv);
	char const *const *p = (char const *const *)(5+argv);
	char const *const *arg0 = q;

    if (0!=strlen(argv[3])) {
	sprintf(new_vscript, "-Wl,--version-script=%s", argv[3]);
	*q++ = cmd;
	*q++ = new_vscript;
    }
    else {
	arg0 = ++q;
	*q++ = cmd;
    }
	for (; plibfn!=(1+p); ++p) {
		*q++ = *p;
	}
	*q = objname;

	/* Re-load with a different version-spec, and with export vector. */
	do_exec(arg0, 0,0,0);
	strcat(tmpname, ".s");
	/*unlink(tmpname);  DEBUGGING */

	/* Save compiled export vector (enables relink using same vector) */
	argv[0] = "/bin/cat";
	argv[1] = objname;
	argv[2] = 0;
	do_exec((char const *const *)(0+argv), 0,(char const *)3,0);
      {	/* Synchronize modification time of text and compiled vector */
	struct stat st;
	if (0==fstat(3, &st)) {
		struct utimbuf ut;
		ut.actime  = st.st_atime;
		ut.modtime = st.st_mtime;
		if (0!=utime("/proc/self/fd/1", &ut)) {
			bad_sync(argv[0]);
		}
	}
	else {
		bad_sync(argv[0]);
	}
      }
	/*unlink(objname);  DEBUGGING */
    }

	free(d_dynsym);
	if (0==rebuild)  {
		free_symbol_patterns((regex_t *)sympats); sympats = 0;
	}
	else {
		free((char **)sympats); sympats = 0;
	}
	free_symbol_patterns((regex_t *)exclude); exclude = 0;

	return 0;
}
