/* impvec.c -- build an import transfer vector for an ELF executable
   Copyright 2000,2001 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <sys/types.h>

#include <elf.h>
#include <stdio.h>  /* fprintf stderr */
#include <stdlib.h>  /* free */
#include <string.h>  /* strncmp */
#include <unistd.h>  /* unlink */

#include "elflib.h"
#include "syslib.h"
#include "veclib.h"

static Elf32_Ehdr const *const *
do_ldd(char const *const fname)
{
	char *const tmpname = my_mktmpname("iv_");
	char const *argv[3];
	argv[0] = "ldd";
	argv[1] = fname;
	argv[2] = 0;
	do_exec(argv, 0,(char const *)mkstemp(tmpname),0);

  {
	Elf32_Ehdr const **ehdrs = 0;
	int n_so = 0;
	size_t sz_ldd = 0;
	char *const ldd_output = getfile(tmpname, &sz_ldd, 1);
	char *p, *const end = sz_ldd + ldd_output;
	unlink(tmpname);

	for (p = ldd_output; p<end; ++p) if ('='==p[0] && '>'==p[1]) {
		++n_so;
	}
	ehdrs = (Elf32_Ehdr const **)zalloc((1+n_so)*sizeof(*ehdrs));
	n_so = 0;
	for (p = ldd_output; p<end; ++p) if ('='==p[0] && '>'==p[1]) {
		char *q;
		for (q=3+p; '\n'!=*q; ++q) ;
		*q=0;
		if (0==strcmp("not found", q-9)) {
			p = q;
			while (ldd_output<=q && '\n'!=*--q) ;
			fprintf(stderr,"warning: %s\n", 1+q);
		}
		else {
			size_t junk;
			while ('('!=*--q);
			*--q=0;
			ehdrs[n_so++] = elf_mmap(3+p, &junk);
		}
	}
	ehdrs[n_so] = 0;
	return ehdrs;
  }
}

struct Import {
	char *image;
	char const *vecname;
	char const *const *symbols;
	int n_symbols;
};

int
main(
	int argc,
	char       *      *const argv
)
{
	/* impvec xclude import_list... loader_args... */
/*
	Load with "-Wl,--noinhibit-exec".  
	Run ldd on the a.out to identify the shared libraries.
	Match shared library to vector.
	For each vector:
		Identify the symbols that are used.
		Construct definition file, assemble.
	Load with import vector.
*/
	regex_t const *const xclude = get_symbol_patterns(argv[1], 0);

	/* identify the files which are import lists */
	struct Import *const imlist = (struct Import *)zalloc(argc*sizeof(struct Import));
	int nImports = 0;
	int j;
	for (j=2; j<argc; ++j) {
		size_t size = 0;
		char *const image = getfile(argv[j], &size, 0);
		if (0!=image && 0==strncmp("__V_", image, 4)) {
			imlist[nImports].image = image;
			imlist[nImports].symbols = 1+
				get_tokens(image, size, &imlist[nImports].n_symbols);
			imlist[nImports].vecname = imlist[nImports].symbols[-1];
			imlist[nImports].n_symbols--;  /* vecname */
			++nImports;
		}
		else { /* end of import lists */
			break;
		}
	}

  {
	/* Load what is there now; force output even if undefined symbols. */
	char **const ld_args = j+argv;
	ld_args[-1] = ld_args[0];
	ld_args[ 0] = "-Wl,--noinhibit-exec";
	do_exec((char const *const *)(-1+ld_args), 0,0,"/dev/null");

   {
	char const *const *const poutfn = find_outfn((char const *const *)ld_args);
	size_t outsz = 0;
	Elf32_Ehdr const *const exe = elf_mmap(*poutfn, &outsz);
	Elf32_Shdr const *const exestrtabsec = 
		elf_find_section_name(".strtab", exe);
	char const *const exestrtab = (char const *)
		(exestrtabsec->sh_offset + (char *)exe);
	Elf32_Shdr const *const exesymtabsec = 
		elf_find_section_name(".symtab", exe);
	Elf32_Sym const *const exesym0 = (Elf32_Sym const *)
		(exesymtabsec->sh_offset + (char *)exe);
	int const n_exesym = exesymtabsec->sh_size / sizeof(Elf32_Sym);

	/* Identify and map the shared libraries. */
	Elf32_Ehdr const *const *const solibs = do_ldd(*poutfn);
	Elf32_Ehdr const *const *ppso = solibs;
	char const *fn_so = "foo.so";  /* NYI: need names from ldd! */
	int n_center = 0;  /* number of previous centers */
	FILE *const impf = stdout;

	/* Foreach shlib, see if it has a vector that matches an import list.
	   If so, then use the import list to satisfy undefineds from exe.
	*/
	for (; 0!=*ppso; ++ppso) {
		Elf32_Ehdr const *const ehdr = *ppso;
		Elf32_Shdr const *const dynsec = 
			elf_find_section_type(SHT_DYNAMIC, ehdr);
		Elf32_Dyn const *const dynamic = (Elf32_Dyn const *)
			(dynsec->sh_offset + (int)ehdr);
		char const *const strtab = (char const *)
			elf_find_dynamic(DT_STRTAB, dynamic, ehdr);
		char const *const so_vecname = get_vecname(strtab, dynamic, fn_so);

		struct Import const *imp = imlist;
		int n_im = nImports;
		for (; --n_im>=0; ++imp) if (0==strcmp(so_vecname, imp->vecname)) {
			char *const needed = zalloc(imp->n_symbols);
			Elf32_Sym const *exesymp = exesym0;
			int j = n_exesym;
			for (; --j>=0; ++exesymp) if (SHN_UNDEF==exesymp->st_shndx) {
				char const *const name = exesymp->st_name + exestrtab;
				if (0==matches(name, xclude)) {
					char const *const *vnp = imp->symbols;
					int k;
					for (k=0; k < imp->n_symbols; ++vnp, ++k)
					if (0==strcmp(name, *vnp)) {
						needed[k] = 1;
					}
				}
			}
		  {
			int n_thisvec = 0;  /* number of symbols used from this vector */
			int d_center = 0;  /* displacement from center feed */

			for (j=0; j < imp->n_symbols; ++j) if (0!=needed[j]) {
				char const *const name = imp->symbols[j];
				if (0==n_thisvec++) { /* first symbol that is used from this vector */
					d_center = 127;
					fprintf(impf,
						"	.data; .align 1<<3; .D%s:\n"
						"	.text; .align 1<<2\n",
							so_vecname );
				}
				fprintf(impf, "\n"
					"	.data; .word %d; .byte 0\n"
					"%s:\n"
					"	jmp    .I%s\n",
						1+j, name, name);
				fprintf(impf,
					"	.text; .I%s: pushl $%d",
					name, n_thisvec - 128 -1 );
				d_center -= (n_thisvec<=256) ? 2 : 5;
				if (d_center < 0) {
					d_center = 127;
					fprintf(impf, "\n.C%d: jmp .C%s\n",
						n_center++, so_vecname);
				}
				else {
					d_center -= 2;
					fprintf(impf, "; jmp .C%d\n", n_center);
				}
				fprintf(impf,
					"	.weak    %s\n"
					"	.type    %s,@function\n",
					name, name);
			}
			if (0!=n_thisvec) { /* terminate list of ordinals */
				fprintf(impf, "\n	.data; .word 0\n\t.text\n");
				if (127 != d_center) {
					fprintf(impf, ".C%d:\n", n_center++);
				}
				fprintf(impf,
					".C%s:\n"
					"	push %%eax; push %%eax  # rest of room for iret\n"
					"	pusha\n"
					"	jmp .E%s\n"
					"	.data\n"
					".E%s:\n"
					"	movl $.D%s,%%edx  # relative relocation here\n"
					"	movl $  %s,%%eax  # external relocation here\n"
					"	jmp .subst\n",
						so_vecname, so_vecname, so_vecname, so_vecname, so_vecname);
			}
		  }
			free(needed);
		}
	}
	fprintf(impf, "%s",
		"	.text\n"
		".subst:\n"
		"	movl $6,%ebx\n"
		"	call *%eax   # get %ebx= src vector\n"
		"	push %edx  # save dst vector\n"
		"	jmp .subst7\n"
		".subst1:\n"
		"	movl (%ebx,%eax,4),%eax  # offset from base\n"
		"	addl %ebx,%eax  # destination\n"
		"	subl %edx,%eax  # displacement\n"
		"	movl %eax,-4(%edx)  # d32 of 'jmp'\n"
		".subst7:\n"
		"	movzwl (%edx),%eax; addl $8,%edx\n"
		"	decl %eax; jns .subst1\n"
		"	popl %edx  # restore dst vector\n"
		"	pushf; popl %eax; xchgl 10*4(%esp),%eax  # index of trigger\n"
		"	leal 8+128*8(%edx,%eax,8),%eax  # base for jmp\n"
		"	addl -4(%eax),%eax  # add d32 of jmp\n"
		"	movl %eax,8*4(%esp)  # dst of trigger\n"
		"	push %cs; popl %eax; movl %eax,9*4(%esp)\n"
		"	popa\n"
		"	iret\n");
	free((void *)solibs);
   }
  }
  {
	free_symbol_patterns((regex_t *)xclude);
	for (j=nImports; --j>=0; ) {
			free((void *)(imlist[j].symbols -1));
	}
	free(imlist);
  }
	return 0;
}
/*
vi: ts=4
*/
