/* recovervec.c -- recover the export vector from a shared library
   Copyright 2000 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <sys/types.h>

#include <elf.h>
#include <stdlib.h>  /* free */
#include <stdio.h>  /* fprintf, stderr, lseek */
#include <string.h>  /* strncmp */

#include "elflib.h"
#include "syslib.h"
#include "veclib.h"

struct x {
	Elf32_Addr value;
	char const *name;
};

static int
cmp_x(void const *pp, void const *qq)  /* qsort comparison fuction */
{
	struct x const *const p = (struct x const *)pp;
	struct x const *const q = (struct x const *)qq;
	if (p->value < q->value) return -1;
	if (p->value > q->value) return  1;
	if (p->name  < q->name)  return -1;
	if (p->name  > q->name)  return  1;
	                         return  0;
}

int
main(
	int argc,
	char       *      *const argv
)
{
	/* map the .so, and locate interesting pieces */
	size_t libsz = 0;
	Elf32_Ehdr const *const ehdr = elf_mmap(argv[1], &libsz);
	Elf32_Shdr const *const sectab = (Elf32_Shdr const *)(ehdr->e_shoff + (char *)ehdr);
	Elf32_Shdr const *const dynsec = elf_find_section_type(SHT_DYNAMIC, ehdr);
	Elf32_Shdr const *const symsec = elf_find_section_type(SHT_DYNSYM, ehdr);
	int const n_dynsym = symsec->sh_size / sizeof(Elf32_Sym);
	Elf32_Dyn const *const dynamic = (Elf32_Dyn const *)(dynsec->sh_offset + (int)ehdr);
	Elf32_Sym const *const dynsym = (Elf32_Sym const *)elf_find_dynamic(DT_SYMTAB, dynamic, ehdr);
	char const *const dynstr = (char const *)elf_find_dynamic(DT_STRTAB, dynamic, ehdr);

	int n_symsym;
	Elf32_Addr base = 0;
	Elf32_Addr *p;
	Elf32_Sym const *vsym = 0;
	Elf32_Sym const *symp = dynsym;
	int n_vec;
	int j;
	for (j=n_dynsym; --j>=0; ) { /* tends to be at the far end */
		char const *const name = symp[j].st_name + dynstr;
		if (0==(strncmp("__V_", name, 4))) {
			vsym = &symp[j];
			base = 11 + symp[j].st_value;
			p = (Elf32_Addr *)(base + (char *)ehdr);
			for (n_vec = 0; *p++; ++n_vec) ;
			break;
		}
	}
	if (0!=vsym) {
		printf("%s\n", vsym->st_name + dynstr);
	}
	else {
		fprintf(stderr, "No __V_ symbol in .dynsym\n");
		exit(1);
	}
	n_symsym = n_dynsym;
	for (j=ehdr->e_shnum; --j>=0; ) if (sectab[j].sh_type==SHT_SYMTAB) {
		n_symsym += sectab[j].sh_size / sectab[j].sh_entsize;
	}
  {
	struct x *const xsym = (struct x *)malloc(sizeof(*xsym) * n_symsym);
	struct x *const xvec = (struct x *)malloc(sizeof(*xvec) * n_vec);
	struct x *xp;
	for (xp=xsym, symp=dynsym, j=0; j<n_dynsym; ++xp, ++symp, ++j) {
		xp->value = symp->st_value;
		xp->name  = symp->st_name + dynstr;
	}
	for (j=ehdr->e_shnum; --j>=0; ) if (sectab[j].sh_type==SHT_SYMTAB) {
		char const *const strtab = sectab[sectab[j].sh_link].sh_offset +
			(char const *)ehdr;
		int k = sectab[j].sh_size / sectab[j].sh_entsize;
		symp = (Elf32_Sym const *)(sectab[j].sh_offset + (char *)ehdr);
		for (; --k>=0; ++symp) {
			unsigned const shindx = symp->st_shndx;
			if (SHN_LORESERVE<=shindx
			||  sectab[shindx].sh_type != SHT_PROGBITS) {
				continue;
			}
			xp->value = symp->st_value;
			xp->name  = symp->st_name + strtab;
			++xp;
		}
	}
	n_symsym = xp - xsym;

	p = (Elf32_Addr *)(base + (char *)ehdr);
	for (j=0; *p; ++j, ++p) {
		xvec[j].value = *p + base;
		xvec[j].name = (char const *)j;
	}
	qsort(xsym, n_symsym, sizeof(*xsym), cmp_x);
	qsort(xvec, n_vec,    sizeof(*xvec), cmp_x);
    {
	char const **const names = (char const **)malloc(sizeof(*names) * n_vec);
	struct x const *psym = xsym;
	struct x const *pvec = xvec;
	for (j=n_vec; 0!=j; ) {
		if (psym->value < pvec->value) {
			++psym; continue;
		}
		if (psym->value > pvec->value) {
			++pvec; continue;
		}
		names[(unsigned)pvec->name] = psym->name;
		++psym; ++pvec; --j;
	}
	for (j=0; j<n_vec; ++j) {
		printf("%s\n", names[j]);
	}
	free(names);
    }
	free(xvec);
	free(xsym);
  }
	return 0;
}

