/* relvec.c -- finalize relocations of export vector and R_386_RELATIVE
   Copyright 2001 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
typedef unsigned size_t;  /* redefined; but system #include<...> are a mess. */

#include "elflib.h"
#include "veclib.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* relvec a.elf incl excl  > b.elf
   incl, excl: optional regex for symbols.  If defined, and match incl but
   not excl, then R_386_32, R_386_GLOB_DAT, and R_386_JMP_SLOT relocs
   on those symbols are finalized, too.
   Typical excl might be:
     ^__ti  because g++ 2.95 uses those for rtti (RunTimeTypeInfo)
            which must be Common across all modules [?]
     ^__environ$  because 3-parameter main() got un-standardized into Common
     ^__ctype_  bash cheats here
   For glibc-2.x (or other shared library with functions commonly overridden),
   typical incl might be '^__ ^_[A-Z]' because ISO C89 reserves those prefixes
   for "internal" names which the client is supposed to avoid, and thus should
   not be overriding, either.
   For a shared library that is designed as a subsystem, with little or no
   usage of function override, then typical incl might be '.' [match every
   symbol].
*/
int
main(int argc, char *const *const argv)
{
	regex_t const *const exclpat = (argc<=3) ? 0 : get_symbol_patterns(argv[3], 0);
	regex_t const *const inclpat = (argc<=2) ? 0 : get_symbol_patterns(argv[2], 0);
	size_t elf_size;
	Elf32_Ehdr const *const ehdr = elf_mmap(argv[1], &elf_size);
	Elf32_Phdr const *const pt_dyn =
		elf_find_phdr_type(PT_DYNAMIC, ehdr);
	Elf32_Dyn *const dynamic = (Elf32_Dyn *)
		(pt_dyn->p_offset + (char *)ehdr);
	Elf32_Shdr const *const shdr0 = (Elf32_Shdr *)
		(ehdr->e_shoff + (int)ehdr);
	Elf32_Shdr *secrel = (Elf32_Shdr *)
		elf_find_section_name(".rel.plt", ehdr);
	char const *const shstrtab =
		elf_find_section_name(".shstrtab", ehdr)->sh_offset +
			(char const *)ehdr;
	size_t const szrel = elf_get_dynamic(DT_RELSZ, dynamic);
	size_t const rp_base = elf_get_dynamic(DT_REL, dynamic);
	Elf32_Rel *const rp0 = (Elf32_Rel *)
		elf_find_dynamic(DT_REL, dynamic, ehdr);
	Elf32_Rel *rp = (Elf32_Rel *)(
		elf_get_dynamic(DT_PLTRELSZ, dynamic) + (char *)
		elf_find_dynamic(DT_JMPREL, dynamic, ehdr) );
	Elf32_Rel *qp = rp;
	Elf32_Sym const *const symtab = (Elf32_Sym const *)
		elf_find_dynamic(DT_SYMTAB, dynamic, ehdr);
	char const *const strtab = (char const *)
		elf_find_dynamic(DT_STRTAB, dynamic, ehdr);
	int need_txtrel = 0;
	int nerr = 0;

	for (; rp0 <= --rp; ) {
		Elf32_Sym const *const symp = ELF32_R_SYM(rp->r_info) + symtab;
		char const *const name = symp->st_name + strtab;
		unsigned const rtype = ELF32_R_TYPE(rp->r_info);
		int *const p = (int *)elf_find_offset(rp->r_offset, ehdr);
		if (SHN_COMMON==symp->st_shndx) {
			fprintf(stderr, "common #%x: %#x %#x %s\n",
				(char *)rp - (char *)ehdr, rp->r_offset, rp->r_info, name);
			goto move;
		} else
		if (R_386_PC32==rtype) {
			if (SHN_UNDEF==symp->st_shndx ) {
				fprintf(stderr,"undef in R_386_PC32 at %#x: %#x %#x %s\n",
					(char *)rp - (char *)ehdr, rp->r_offset, rp->r_info, name);
				{
					Elf32_Rel const tmp = *--qp;
					*qp = *rp;
					*rp = tmp;
				}
			}
			else {
				*p += symp->st_value - rp->r_offset;
			}
		}
		else if (R_386_RELATIVE==rtype) {
			/* nothing to do because we assume it does not move */
		}
#if 1  /*{ "logically correct" except for C++ "one definition rule" etc. [?] */
		else if (( R_386_32==rtype
		  || R_386_GLOB_DAT==rtype
		  || R_386_JMP_SLOT==rtype )
		&& (0!=inclpat &&  matches(name, inclpat))
		&& (0==exclpat || !matches(name, exclpat)) ) switch (rtype) {
		case R_386_GLOB_DAT:
			/* fall through */
		case R_386_32: {
			if (SHN_UNDEF!=symp->st_shndx) {
				*p += symp->st_value;
			}
			else {
				goto move;
			}
		} break;
		case R_386_JMP_SLOT: {
			if (SHN_UNDEF!=symp->st_shndx) {
				*p  = symp->st_value;  /* no '+' */
			}
			goto move;  /* never compress JMPREL relocs */
		} break;
		}
#endif  /*}*/
		else {
move:
			{
				Elf32_Rel const tmp = *--qp;
				*qp = *rp;
				*rp = tmp;
			}
		}
		if (((int)rp - (int)ehdr)==secrel->sh_offset) {
			if (0!=strcmp(".rel.plt", secrel->sh_name + shstrtab)) {
				int const d = (int)qp - (int)rp;
				secrel->sh_addr += d;
				secrel->sh_offset += d;
				secrel->sh_size = (1+secrel)->sh_addr - secrel->sh_addr;
				if (0!=secrel->sh_size
				&&  0==(SHF_WRITE & shdr0[secrel->sh_info].sh_flags) ) {
					++need_txtrel;
				}
			}
			--secrel;
		}
	}
	++rp;  /* 'for' loop overran the beginning */

	{
		int const d = (int)qp - (int)rp;
		elf_set_dynamic(DT_RELSZ, szrel - d, dynamic);
		elf_set_dynamic(DT_REL, rp_base + d, dynamic);
	}
	if (0!=need_txtrel) {
		fprintf(stderr, "warning: still need DT_TXTREL\n");
	}
	else {
		Elf32_Dyn const *dp_relent = 0;
		Elf32_Dyn       *dp_txtrel = 0;
		unsigned t = pt_dyn->p_filesz;
		Elf32_Dyn *dp = dynamic;
		for (; 0 < t; ++dp, t-=sizeof(*dp)) {
			if (DT_RELENT ==dp->d_tag
			||  DT_RELAENT==dp->d_tag ) {
				dp_relent = dp;
			}
			else if (DT_TEXTREL==dp->d_tag) {
				dp_txtrel = dp;
			}
		}
		if (0!=dp_txtrel) {
			if (0!=dp_relent) {
				*dp_txtrel = *dp_relent;
			}
			else {
				fprintf(stderr, "warning: DT_TEXTREL not removed\n");
			}
		}
	}
	return (0!=nerr) || (elf_size!=write(1, ehdr, elf_size));
}

