/* retroimp.c -- retrofit use of vectors into an already linked a.elf.
   Copyright 2000,2001 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <sys/types.h>

#include <elf.h>
#include <stdio.h>  /* fprintf stderr */
#include <stdlib.h>  /* free */
#include <string.h>  /* strncmp */

#include "elflib.h"
#include "syslib.h"
#include "veclib.h"

#define EAX 0
#define ECX 1
#define EDX 2
#define EBX 3
#define ESP 4
#define EBP 5
#define ESI 6
#define EDI 7

#define rAL 0
#define rCL 1
#define rDL 2
#define rBL 3
#define rAH 4
#define rCH 5
#define rDH 6
#define rBH 7

#define ALU(op,mod) (((op)<<3)|(mod))
#  define ADD 0
#  define OR  1
#  define ADC 2
#  define SBB 3
#  define AND 4
#  define SUB 5
#  define XOR 6
#  define CMP 7

#  define B2M 0
#  define W2M 1
#  define B2R 2
#  define W2R 3
#  define B2A 4
#  define W2A 5

#define CALL  0xE8
#define FETCH 0x8B
#define IMMSBL 0x83
#define JMP8  0xEB
#define JMP32 0xE9
#define JMPI  4
#define LDA   0xA1
#define LEA   0x8D
#define MOVIR 0xB8
#define MOVSBL 0xBE
#define MOVSWL 0xBF
#define MOVZBL 0xB6
#define MOVZWL 0xB7
#define NEG   3
#define NOP   0x90
#define POP   0x8F
#define POPA  0x61
#define POPR  0x58
#define PUSHA 0x60
#define PUSHI8 0x6A
#define PUSHR 0x50
#define SHIFTI 0xC1
#  define ROL 0
#  define ROR 1
#  define RCL 2
#  define RCR 3
#  define SHL 4
#  define SHR 5
#  define SAL SHL
#  define SAR 7
#define STA   0xA3
#define STORE 0x89
#define TEST  0x85

#define JO  0x0
#define JNO 0x1
#define JC  0x2
#define JNC 0x3
#define JZ  0x4
#define JNZ 0x5
#define JBE 0x6
#define JA  0x7
#define JS  0x8
#define JNS 0x9
#define JP  0xA
#define JNP 0xB
#define JL  0xC
#define JGE 0xD
#define JLE 0xE
#define JG  0xF

#define D0 0
#define D8 1
#define D32 2
#define RR 3

#define S1 0
#define S2 1
#define S4 2
#define S8 3

#define MAB(m,a,b) (((m)<<6) | ((a)<<3) | (b))
#define SIB(s,i,b) (((s)<<6) | ((i)<<3) | (b))

struct Import {
	char *image;
	char const *vecname;
	char const *const *symbols;
	int n_symbols;
	unsigned char used;
	  signed char vecn;
};

struct plt86 {
	unsigned char jmpi[6];
	unsigned char pushi[5];
	unsigned char jmpl[5];
};

struct vredir {
	  signed char vec;  /* which vector */
	unsigned char gap;  /* distance to next vredir */
	unsigned short ord;  /* symbol ordinal in vector */
};

static void
nop(unsigned char *p, int n)
{
	unsigned r = EDI;
	while (7<=n) {
		*p++ = LEA; *p++ = MAB(D32,r,ESP); *p++ = SIB(S1,ESP,r);
		*(int *)p = 0; p += 4;
		r ^= 1; n -= 7;
	}
	while (6<=n) {
		*p++ = LEA; *p++ = MAB(D32,r,r);
		*(int *)p = 0; p += 4;
		r ^= 1; n -= 6;
	}
	while (3<=n) {
		*p++ = LEA; *p++ = MAB(D8,r,r); *p++ = 0;
		r ^= 1; n -= 3;
	}
	while (2<=n) {
		*p++ = FETCH; *p++ = MAB(RR,r,r);
		r ^= 1; n -= 2;
	}
	while (1<=n) {
		*p++ = NOP;
		n -= 1;
	}
}

static unsigned
round_up(unsigned u, unsigned m)
{
	return ((m -1 + u) / m) * m;
}

static unsigned
strtablen(Elf32_Sym const *symp, int nsym)
{
	unsigned len = 0;
	for (; --nsym>=0; ++symp) {
		len += 1+ strlen((char const *)symp->st_name);
	}
	return len;
}

int
main(
	int argc,
	char       *      *const argv
)
{
	/* retroimp xclude import_list... a.elf  > a_vec.elf */
/*
	Structure of existing plt86 within PLT (Program Linkage Table) is
			jmp *[pFoo: .long iFoo]  # "[]" denotes literal table
		iFoo:
			pushl $ displ_within_.rel.plt_of_reloc_for_pFoo
			jmp PLT  # tail merge
	and the PLT fixup changes from ".long iFoo" to ".long Foo".

	Change to
			jmp *[pFoo: .word ordinal_in_vec; .byte which_vector, d_next]
		{10 bytes to use for code which does runtime relocation}
	and change entry point of a.elf to process vector entries first.
*/
	regex_t const *const xclude = get_symbol_patterns(argv[1], 0);
	Elf32_Sym *const vecsym = (Elf32_Sym *)malloc(sizeof(Elf32_Sym) * argc);

	/* identify the files which are import lists */
	struct Import *const imlist = (struct Import *)zalloc(argc*sizeof(struct Import));
	int nImports = 0;
	int j;
	for (j=2; j<argc; ++j) {
		size_t size = 0;
		char *const image = getfile(argv[j], &size, 0);
		if (0!=image && 0==strncmp("__V_", image, 4)) {
			imlist[nImports].image = image;
			imlist[nImports].symbols = 1+
				get_tokens(image, size, &imlist[nImports].n_symbols);
			imlist[nImports].vecname = imlist[nImports].symbols[-1];
			imlist[nImports].n_symbols--;  /* vecname */
			++nImports;
		}
		else { /* end of import lists */
			break;
		}
	}
	fprintf(stderr,"%d import vectors\n", nImports);
	if (0==nImports) {
		exit(1);
	}

  {
	int nUsedVec, nUsedPLT = 0;
	size_t sz_aelf = 0;
	Elf32_Ehdr *const ehdr = elf_mmap(argv[argc -1], &sz_aelf);
	Elf32_Phdr const *const pt_dynamic = elf_find_phdr_type(PT_DYNAMIC, ehdr);
	Elf32_Phdr const *const pt_text = elf_find_phdr_type(PT_LOAD, ehdr);
	unsigned const tx_reloc = pt_text->p_vaddr - pt_text->p_offset;
	Elf32_Dyn *const dynamic = (Elf32_Dyn *)(pt_dynamic->p_offset + (char *)ehdr);
	Elf32_Sym const *const dynsym =
		(Elf32_Sym const *)elf_find_dynamic(DT_SYMTAB, dynamic, ehdr);
	char const *const dynstr =
		(char const *)elf_find_dynamic(DT_STRTAB, dynamic, ehdr);
	/*unsigned *const plt = (unsigned *)(
		elf_find_offset(elf_get_dynamic(DT_PLTGOT, dynamic), ehdr);*/
    int const pltrelsz = elf_get_dynamic(DT_PLTRELSZ, dynamic);
	Elf32_Rel *const pltrel = (Elf32_Rel *)elf_find_dynamic(DT_JMPREL, dynamic, ehdr);
	Elf32_Shdr const *const symsec = elf_find_section_type(SHT_DYNSYM, ehdr);
	int const n_dynsym = symsec->sh_size / sizeof(Elf32_Sym);

	Elf32_Rel const *const rp_end = (Elf32_Rel const *)(pltrelsz + (char *)pltrel);
	Elf32_Rel const *rp = pltrel;
	unsigned char *head_icept = 0;
	struct vredir *prev_vredir = 0;
	for (; rp < rp_end; ++rp) {
		Elf32_Sym const *const symp = ELF32_R_SYM(rp->r_info) + dynsym;
		char const *const symname = symp->st_name + dynstr;
		struct vredir *const vp = (struct vredir *)elf_find_offset(rp->r_offset,
			ehdr);
		unsigned char *const icept = (unsigned char *)
			 elf_find_offset(*(unsigned *)vp, ehdr);
		int i;
		if (R_386_JMP_SLOT!=ELF32_R_TYPE(rp->r_info)) {
			fprintf(stderr,"not R_386_JMP_SLOT: %d\n", ELF32_R_TYPE(rp->r_info));
			exit(1);
		}
		if ((Elf32_Rel const *)(*(int *)(1+icept) + (char *)pltrel) != rp) {
			fprintf(stderr,"mismatched rp #%d\n", rp - pltrel);
			exit(1);
		}
		if (matches(symname, xclude)) {
			fprintf(stderr, "excluded: %s\n", symname);
		}
		else for (i=0; i<nImports; ++i) {
			char const *const *vecsym = imlist[i].symbols;
			int j;
			for (j=0; j<imlist[i].n_symbols; ++vecsym, ++j) {
				if (0==strcmp(symname, *vecsym)) {
					imlist[i].used = 1; ++nUsedPLT;
					fprintf(stderr, "%s: %s\n", symname, imlist[i].vecname);
					if (prev_vredir) {
						unsigned const t = vp - prev_vredir;
						/*fprintf(stderr, "gap = %d\n", t);*/
						if (255 < t) {
							fprintf(stderr, "gap too big: %d\n", t);
							exit(1);
						}
						prev_vredir->gap = t;
					}
					vp->ord = j;
					vp->vec = i;
					vp->gap = 0;
					if (0==head_icept) {
						head_icept = icept;
					}
					prev_vredir = vp;
					goto next_rp;
				}
			}
		}
next_rp: ;
	}
	free_symbol_patterns((regex_t *)xclude);

  {/* dribble code into plt which were converted to use vector */
	int j,t;
	unsigned char *dst, *src = 0, *src2;
	unsigned char *p = head_icept;
	unsigned const e_entry = ehdr->e_entry;
	Elf32_Rel *rp = (Elf32_Rel *)(*(int *)(1+ p) + (char *)pltrel);
	unsigned const rp0_offset = rp->r_offset;
	struct vredir *const vp0 = (struct vredir *)elf_find_offset(rp->r_offset,
		ehdr);
	struct vredir *vp = vp0;

	for (nUsedVec = 0, j=0; j<nImports; ++j) if (0!=imlist[j].used) {
		imlist[j].vecn = -(1 + nUsedVec++);
	}
	fprintf(stderr, "%d used vectors\n", nUsedVec);
	if (31 < nUsedVec) {
		fprintf(stderr, "31 is the maximum number of used vectors\n");
		exit(1);
	}
	fprintf(stderr, "%d used slots in PLT\n", nUsedPLT);
	if (nUsedPLT < (13 + 5*nUsedVec)) {
		fprintf(stderr, "need %d used slots in PLT, but have only %d\n",
			13 + 5*nUsedVec, nUsedPLT);
		exit(1);
	}
	for (nUsedVec = 0, j=0; j<nImports; ++j) if (0!=imlist[j].used) {
		Elf32_Rel     *const rp_save = rp;
		struct vredir *const vp_save = vp;
		vecsym[nUsedVec].st_name = (unsigned)imlist[j].vecname;
		vecsym[nUsedVec].st_value = dynsym[ELF32_R_SYM(rp->r_info)].st_value;
		vecsym[nUsedVec].st_size = 0;
		vecsym[nUsedVec].st_info = ELF32_ST_INFO(STB_GLOBAL, STT_FUNC);
		vecsym[nUsedVec].st_other = 0;
		vecsym[nUsedVec].st_shndx = SHN_UNDEF;

		dst = p; rp->r_info = ELF32_R_INFO(nUsedVec + n_dynsym,
			ELF32_R_TYPE(rp->r_info) );

		vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
		/* *((unsigned *)(1+p)) = sizeof(Elf32_Rel) * nUsedVec + pltrelsz; */
		p += t; rp += vp->gap; vp += vp->gap;

		vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
		if (src) {
			((int *)src)[-1] = p - src; src = 0;
		}
		if (0==nUsedVec) {
			ehdr->e_entry = tx_reloc + ((char *)p - (char *)ehdr);
			p[0] = PUSHA;
		}
		else {
			p[0] = PUSHR+EBX;
		}
		p[1] = PUSHI8; p[2] = 6; p[3] = POPR+EBX;
		p[4] = JMP32; *(int *)(5+p) = t - (4+5);
		nop(9+p, 1);
		p += t; rp += vp->gap; vp += vp->gap;

		vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
		p[0] = CALL;  *(int *)(1+p) = dst - (5+p);
		p[5] = JMP32; *(int *)(6+p) = t - (5+5);
		nop(10+p, 0);
		p += t; rp += vp->gap; vp += vp->gap;

		vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
		p[0] = MOVIR+EAX; *(int *)(1+p) = *(unsigned *)vp_save;
		p[5] = JMP32;     *(int *)(6+p) = t - (5+5);
		nop(10+p, 0);
		*(unsigned *)vp_save = 6+ vecsym[nUsedVec].st_value;
		p += t; rp += vp->gap; vp += vp->gap;

		vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
		p[0] = STA;   *(int *)(1+p) = rp_save->r_offset;
		p[5] = JMP32; src = (5+5) + p;
		nop(10+p, 0);
		p += t; rp += vp->gap; vp += vp->gap;

		++nUsedVec;
	}

	vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
	if (src) {
		((int *)src)[-1] = p - src; src = 0;
	}
	p[0] = PUSHR+EBX;
	p[1] = LEA; p[2] = MAB(D8,ESI,ESP); p[3] = SIB(S1,ESP,ESP); p[4] = 4*nUsedVec;
	p[5] = JMP32; *(int *)(6+p) = t - (5+5);
	nop(10+p, 0);
	p += t; vp += vp->gap;

	vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
	p[0] = MOVIR+EDX; *((int *)(1+p)) = rp0_offset;
	p[5] = JMP32; *(int *)(6+p) = t - (5+5);
	nop(10+p, 0);
	p += t; vp += vp->gap;

	dst = p;
	{
		static
		struct instr {
			unsigned char i_len;
			unsigned char bytes[3];
		} const subst[] = { /* grouped as if 2-byte JMP is used */
			{2, {FETCH, MAB(D0,EAX,EDX)}},  /* %eax= control word */
			{3, {0x0F, MOVSBL, MAB(RR,EBX,rAL)}},  /* %ebx= vec# */
			{3, {0x0F, MOVZBL, MAB(RR,ECX,rAH)}},  /* %ecx= gap  */

			{3, {FETCH, MAB(D0,EBX,ESP), SIB(S4,EBX,ESI)}},  /* %ebx= &vec */
			{3, {SHIFTI, MAB(RR,SHR,EAX), 16}},  /* %eax= ordinal */

			{3, {FETCH, MAB(D0,EAX,ESP), SIB(S4,EAX,EBX)}},  /* %eax= displ */
			{2, {ALU(ADD,W2R), MAB(RR,EAX,EBX)}},  /* %eax= target */
			{2, {STORE, MAB(D0,EAX,EDX)}},  /* dst */

			{3, {LEA, MAB(D0,EDX,ESP), SIB(S4,ECX,EDX)}},  /* %edx= next */
			{2, {0xF7, MAB(RR,NEG,ECX)}},  /* Carry=0 iff 0==gap */
			{2, {ALU(SBB,W2R), MAB(RR,EAX,EAX)}},  /* %eax= ~0 iff 0!=gap */

			{5, {ALU(AND,W2A)}}, /* loop - done */

			{5, {ALU(ADD,W2A)}}, /* done */

			{~0}
		};
		struct instr const *ip = subst;
		int room = 8;
		goto start;
		for (;;) {
			if ((t - 2)<=127 && room < ip->i_len) {
				t-=2; p[0]= JMP8; p[1] = t;
	next:
				nop(p+=2, room);
				p += t; vp += vp->gap; room = 8;
	start:
				vp->vec = imlist[vp->vec].vecn; t = vp->gap * sizeof(struct plt86);
				if (15 < ip->i_len) {
					break;
				}
			}
			else if ((room -3) < ip->i_len) {
				t-=5; p[0]= JMP32; *(int *)(1+p) = t;
				p+=3; room-=3;
				goto next;
			}
			else { /* it fits */
				memcpy(p, ip->bytes, ip->i_len);
				p += ip->i_len; t -= ip->i_len; room -= ip->i_len;
				if (5<=ip->i_len) { /* save position for later fixup */
					src2 = src; src = p;
				}
				++ip;
			}
		}
	}
	/* 10 bytes is not enough for a long conditional branch (6 bytes) to 'loop'
	   and a long unconditional branch (5 bytes) to 'done'.  So the address
	   of the destination has been computed using the Carry flag trick and
	   the 5-byte immediate-to-EAX instructions.  Fix up the 4-byte constants.
	*/
	((int *)src )[-1] = tx_reloc + ((char *)p - (char *)ehdr);  /* 'done' */
	((int *)src2)[-1] = dst - p;  /* 'loop' - 'done' */
	src[0] = 0xFF; src[1] = MAB(RR,JMPI,EAX);
	nop(2+src, 3);

	p[0] = FETCH; p[1] = MAB(RR,ESP,ESI);
	p[2] = POPA;
	p[3] = JMP32; *(int *)(4+p) = elf_find_offset(e_entry, ehdr) - (char *)(8+p);
	nop(9+p, 1);

	do {
		vp += vp->gap;
		vp->vec = imlist[vp->vec].vecn;
	} while (vp->gap);
  }
  {
	off_t const versec = elf_get_dynamic(DT_VERSYM, dynamic);
	unsigned *const htab = (unsigned *)elf_find_dynamic(DT_HASH, dynamic, ehdr);
	unsigned const nbucket = htab[0];
	unsigned const nchain  = htab[1];
	unsigned const overhead = round_up(nUsedVec * (
				  sizeof(unsigned)  /* more 'chain' */
				+ sizeof(Elf32_Sym)  /* more symbols */
				+ ((~0!=versec) ? sizeof(short) : 0) )  /* more version */
			+ strtablen(vecsym, nUsedVec),  /* names of vectors */
		sizeof(unsigned));
	unsigned const new_nbucket = nbucket - overhead/sizeof(unsigned);
	unsigned const new_nchain = nUsedVec + nchain;
	unsigned *const new_htab = (unsigned *)zalloc(sizeof(unsigned) *
		(2 + new_nbucket + new_nchain));
	new_htab[0] = new_nbucket;
	new_htab[1] = new_nchain;
	fprintf(stderr,"retroimp: number of buckets  %d ==> %d  (for %d names)\n",
		nbucket, new_nbucket, n_dynsym);
	if (nbucket < new_nbucket) {
		exit(1);
	}
	elf_rehash(new_htab, htab, dynsym, dynstr);
	for (j=0; j<nUsedVec; ++j) {
		elf_insert_dynsym((char const *)vecsym[j].st_name, j+n_dynsym, new_htab);
	}
	veclib_retrofit(1, ehdr, 0, 0, new_htab, vecsym, nUsedVec);
	free(new_htab);
	free(vecsym);
  }
	for (j=nImports; --j>=0; ) {
			free((void *)(imlist[j].symbols -1));
	}
	free(imlist);
  }
	return 0;
}

/*
vi: ts=4
*/
