/* retrovec.c -- retrofit an export vector into an existing shared library
   Copyright 2000 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <sys/types.h>

#include <elf.h>
#include <stdlib.h>  /* free */
#include <stdio.h>  /* fprintf, stderr, lseek */
#include <string.h>  /* strncmp */

#include "elflib.h"
#include "syslib.h"
#include "veclib.h"

#define strneq(a,b) strcmp((a),(b))

static int  /* return symbol ordinal */
find_dynsym(
	char const *const name,
	unsigned const *const htab,
	Elf32_Sym const *const symtab,
	char const *const dynstr
)
{
	unsigned int const nbucket = htab[0];
	/*unsigned int const nchain  = htab[1];*/
	unsigned int const *const buckets = &htab[2];
	unsigned int const *const chains  = &buckets[nbucket];
	unsigned int const m = elf_hash_dynsym(name) % nbucket;
	unsigned int si;

	for (si = buckets[m]; 0!=si; si = chains[si]) {
		char const *const p = symtab[si].st_name + dynstr;
		if (0==strneq(name, p)) {
			return si;
		}
	}
	return 0;
}

static unsigned
round_up(unsigned u, unsigned m)
{
	return ((m -1 + u) / m) * m;
}

struct Import {
	char *image;
	char const *vecname;
	char const *const *names;
	int n_names;
	unsigned *values;
};

unsigned char code_vec[16]  = {
	0xbb, 0,0,0,0,     /* movl $ _GLOBAL_OFFSET_TABLE_ - (5 + <call>),%ebx */
	0xe8, 2,0,0,0,     /* +5: call <addl> */
	0xc3,              /* ret */
	0x90,              /* nop */
	0x03, 0034, 0044,  /* addl (%esp),%ebx */
	0xc3               /* ret */
};

int
main(
	int argc,
	char       *      *const argv
)
{
	/*  retrovec import.vec old.so  > new.so   # new.so must support lseek  */
/*
  Steal the space from the buckets (heads of separate chaining) of the
  .hash DT_HASH table.  Put the Code_vec and table of displacements before
  the .hash table.  Add the symbol for the vector name to the end of
  the .dynsym DT_SYMTAB section.  Add the characters for the vector name
  to the end of the .dynstr DT_STRTAB section.  Add one version number
  to the end of the SHT_GNU_versym section.  Fixup the Elf32_Shdr and
  the .dynamic DT_DYNAMIC section.  Extend DT_STRSZ to cover the whole
  .dynstr section; binutils-2.9.5.0.37 uses the high-address end of the
  .dynstr section to hold the version strings, but does not include them
  in DT_STRSZ.  Avoid touching SHT_GNU_verdef or SHT_GNU_verneed; these
  are horrible data structures that suffer from the Apollo Computer Inc
  data structure disease (intermixed structures of different types and
  sizes, with variable repetition counts), and are an affront to the
  arrays of Elf.  They also contain embedded offsets to .dynstr, which
  would have to be relocated.

  Stealing space from the buckets makes a search take longer.  By default,
  /bin/ld seems to pick 0.6 <= nbuckets/nsymbols <= 1.0, which is generous.
  Be concerned if the ratio drops below 0.3 or so.
*/

	struct Import vector;
  {
	/* get the vector, which is in the format of an import vector:
	   file begins with __V_<name>, followed by list of literal names
	*/
	size_t size = 0;
	char *const image = getfile(argv[1], &size, 0);
	if (0!=image && 0==strncmp("__V_", image, 4)) {
		vector.image = image;
		vector.n_names = 0;
		vector.names = 1+
			get_tokens(image, size, &vector.n_names);
		vector.vecname = vector.names[-1];
		vector.n_names--;  /* vecname */
		vector.values = (unsigned *)(sizeof(code_vec) + zalloc(sizeof(code_vec)
			+ sizeof(unsigned) * (1+ vector.n_names) )); /* 1+: 0 at end */
	}
	else {
		fprintf(stderr,"argv[1] is not an import vector\n");
		exit(1);
	}
  }
  {
	/* map the .so, and locate interesting pieces */
	size_t libsz = 0;
	Elf32_Ehdr       *const ehdr = elf_mmap(argv[2], &libsz);
	Elf32_Shdr const *const dynsec = elf_find_section_type(SHT_DYNAMIC, ehdr);
	Elf32_Shdr const *const symsec = elf_find_section_type(SHT_DYNSYM, ehdr);
	Elf32_Dyn *const dynamic = (Elf32_Dyn *)(dynsec->sh_offset + (int)ehdr);
	int const n_dynsym = symsec->sh_size / sizeof(Elf32_Sym);
	off_t const versec = elf_get_dynamic(DT_VERSYM, dynamic);
	Elf32_Sym *const dynsym = (Elf32_Sym *)elf_find_dynamic(DT_SYMTAB, dynamic, ehdr);
	char const *const dynstr = (char const *)elf_find_dynamic(DT_STRTAB, dynamic, ehdr);
	unsigned *const htab = (unsigned *)elf_find_dynamic(DT_HASH, dynamic, ehdr);
	unsigned const nbucket = htab[0];
	unsigned const nchain  = htab[1];

	unsigned const overhead = round_up(sizeof(code_vec)
			+ sizeof(unsigned)  /* one more 'chain' */
			+ sizeof(Elf32_Sym)  /* one more symbol */
			+ 1+strlen(vector.vecname)  /* name of vector */
			+ ((~0!=versec) ? sizeof(short) : 0),  /* version of vector */
		sizeof(unsigned));
	unsigned const new_nbucket = nbucket - overhead/sizeof(unsigned)
		- (1+ vector.n_names);
	unsigned const new_nchain = 1 + nchain;
	fprintf(stderr,"retrovec: number of buckets  %d ==> %d  (for %d names)\n",
		nbucket, new_nbucket, n_dynsym);
	if (nbucket < new_nbucket) {
		exit(1);
	}
	elf_set_dynamic(DT_STRSZ, symsec->sh_size, dynamic);
  {
	unsigned *const new_htab = (unsigned *)zalloc(sizeof(unsigned) *
		(2 + new_nbucket + new_nchain));
	int j;
	for (vector.values[j=vector.n_names]=0; --j>=0; ) {
		char const *const name = vector.names[j];
		int k = find_dynsym(name, htab, dynsym, dynstr);
		if (0==k) {
			fprintf(stderr,"warning: symbol not found: %s\n", name);
		}
		else {
			static char const *const name_bind[] = {
				"local", "global", "weak"
			};
			static char const *const name_type[] = {
				"notype", "object", "func", "section", "file"
			};
			int const bind = ELF32_ST_BIND(dynsym[k].st_info);
			int const type = ELF32_ST_TYPE(dynsym[k].st_info);
			if (STB_GLOBAL!=bind) {
				fprintf(stderr,"warning: symbol is %s: %s\n",
					((bind < STB_NUM) ? name_bind[bind] : "unknown binding"),
					name );
			}
			if (STT_FUNC!=type) {
				fprintf(stderr,"warning: symbol is %s: %s\n",
					((type < STT_NUM) ? name_type[type] : "unknown type"),
					name );
			}
		}
		vector.values[j] = dynsym[k].st_value;
	}
	new_htab[0] = new_nbucket;
	new_htab[1] = 1 + n_dynsym;
	elf_insert_dynsym(vector.vecname, n_dynsym, new_htab);
	elf_rehash(new_htab, htab, dynsym, dynstr);

#if 0  /*{ versioned symbols trigger false positive diagnostics */
	for (j=nbucket; --j>=0; ) { /* verify the transfer */
			int k;
		for (k=bucket[j]; 0!=k; k=chain[k]) {
			int const t = find_dynsym(dynsym[k].st_name+dynstr,
				new_htab, dynsym, dynstr);
			if (t!=k) {
				fprintf(stderr,"lookup error in new htab %d %d %s\n",
					k,t,dynsym[k].st_name+dynstr);
			}
		}
	}
#endif  /*}*/
	{
		unsigned const base = 16 + elf_get_dynamic(DT_HASH, dynamic);
		char *const ante_hash = -sizeof(code_vec) + (char *)vector.values;
		size_t const sz_ante_hash = sizeof(code_vec) + sizeof(unsigned) *
			(1+ vector.n_names);
		Elf32_Sym vecsym;
		int j;
		vecsym.st_name = (unsigned)vector.vecname;
		vecsym.st_value = -11 + base;
		vecsym.st_size = sizeof(code_vec) - 5;
		vecsym.st_info = ELF32_ST_INFO(STB_GLOBAL, STT_FUNC);
		vecsym.st_other = 0;
		vecsym.st_shndx = 1;
		*(int *)(1+(char *)&code_vec) =
			elf_get_dynamic(DT_PLTGOT, dynamic)  - base;
		for (j=0; j<vector.n_names; ++j) {
			vector.values[j] -= base;
		}
		memcpy(ante_hash, &code_vec, sizeof(code_vec));
		veclib_retrofit(1, ehdr, ante_hash, sz_ante_hash, new_htab, &vecsym, 1);
		free(new_htab);
		free(ante_hash);
	}
   }
  }
	free((void *)(vector.names -1));
	return 0;
}

/*
vi: ts=4
*/
