/* syslib.c -- OS and C utility functions
   Copyright 2000 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <sys/mman.h>  /* mmap */
#include <sys/stat.h>  /* fstat */
#include <sys/wait.h>  /* waitpid */
#include <fcntl.h>  /* open */
#include <stdio.h>  /* perror */
#include <stdlib.h>  /* malloc */
#include <string.h>  /* strncmp */
#include <unistd.h>  /* fork, exec*, read */

#include "syslib.h"  /* prototypes for this file */

char *
getfile(
	char const *const fname,
	size_t *p_size,
	int exit_on_fail
)
{
	char *retval;
	struct stat statbuf;
	int const fd = open(fname, O_RDONLY, 0);

	if (0 > fd) {
		goto fail;
	}
	if (0 > fstat(fd, &statbuf)) {
		goto fail;
	}
	*p_size = statbuf.st_size;
	retval = (char *)mmap(0, statbuf.st_size, PROT_READ|PROT_WRITE,
		MAP_PRIVATE, fd, 0);
	if (((caddr_t)-1) == (caddr_t)retval) {
		int n_read;
		retval = (char *)mmap(0, statbuf.st_size, PROT_READ|PROT_WRITE,
			MAP_ANONYMOUS, 0, 0);
		if (((caddr_t)-1) == (caddr_t)retval) {
			goto fail;
		}
		n_read = read(fd, (caddr_t)retval, statbuf.st_size);
		if (n_read != statbuf.st_size) {
			goto fail;
		}
	}
	close(fd);
	return retval;
fail:
	if (exit_on_fail) {
		perror(fname);
		exit(1);
	}
	return 0;
}

static int
redir(int fd, char const *const fname, int direction)
{
	unsigned t = (unsigned)fname;
	if (0==fname) {
		return ~0;
	}

	if (FOPEN_MAX < t) {
		t = open(fname, direction, 0700);
	}
	if (t != fd) {
		close(fd);
		dup2(t, fd);
		close(t);
	}
	return fd;
}

int
do_exec(
	char const *const *const argv,
	char const *const redir0,
	char const *const redir1,
	char const *const redir2
)
{
	pid_t const pid = fork();
	if (~0==pid) {
		perror("cannot fork");
		exit(1);
	}
	else if (0==pid) { /* child */
		char const *const *p = argv;
		fprintf(stderr, "exec:");
		/*fprintf(stderr, " PATH=%s", getenv("PATH"));*/
		for (; *p; ++p) {
			fprintf(stderr," %s", *p);
		}
		fprintf(stderr, "\n");
		redir(0, redir0, O_RDONLY);
		redir(1, redir1, O_CREAT|O_RDWR);
		redir(2, redir2, O_CREAT|O_RDWR);
		execvp(argv[0], (char *const *)argv);  /* execvp bad proto */
		perror("cannot exec ld");
		exit(1);
	}
	else { /* parent */
		int status = 0;
		pid_t const w = waitpid(pid, &status, 0);
		if (w != pid) {
			return ~0;
		}
		if (WIFEXITED(status)) { /* "normal" exit */
			status = WEXITSTATUS(status);
			if (0==status) { /* good */
				return 0;
			}
			exit(status);
		}
		else if (WIFSIGNALED(status)) { /* caught a signal */
			status = WTERMSIG(status);
			exit(-status);
		}
	}
	return 0;
}

char *
my_mktmpname(char const *const base)
{
	static char tmpname[FILENAME_MAX];
	char const *tmpdir = getenv("TMPDIR");
	if (0==tmpdir) {
		tmpdir = "/tmp";
	}
	sprintf(tmpname, "%s%s%sXXXXXX", tmpdir,
		(('/'!=tmpdir[strlen(tmpdir)-1]) ? "/" : ""), base);
	return tmpname;
}

off_t
ulseek(int fd, off_t offset, int whence)
{
	off_t const s = lseek(fd, offset, whence);
	if (~0==s) {
		perror("lseek");
		exit(1);
	}
	return s;
}

size_t
uwrite(int fd, void const *buffer, size_t length)
{
	size_t const w = write(fd, buffer, length);
	if (w!=length) {
		perror("write");
		exit(1);
	}
	return w;
}


char *
zalloc(size_t size)
{
	char *const p = malloc(size);
	if (0==size) {
		perror("cannot malloc");
		exit(1);
	}
	memset(p, 0, size);
	return p;
}

