/* veclib.c -- utility functions for transfer vectors
   Copyright 2000,2001 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <stdio.h>  /* FILENAME_MAX */
#include <string.h>  /* strcat, strcpy, strrchr */
#include <stdlib.h>  /* free */

#include "elflib.h"
#include "syslib.h"
#include "veclib.h"

char const *const *
find_outfn(char const *const *const argv)
{
	char const *const *p;
	for (p = argv; 0!=*p && 0!=strcmp("-o",*p); ++p) {
		;  /* empty */
	}
	if (0!=*p) {
		return 1+p;
	}
	else {
		static char const *const a_out = "a.out";
		return &a_out;
	}
}

#include <ctype.h>

char const *
get_vecname(
	char const *const strtab,
	Elf32_Dyn const *const dynamic,
	char const *const libfn
)
{
	static char vecname[FILENAME_MAX];
	int const so_ndx = elf_get_dynamic(DT_SONAME, dynamic);
	char const *soname;
	char const *basename;

	if (~0 != so_ndx) {
		soname = so_ndx + strtab;
	}
	else {
		soname = libfn;
	}
	basename = strrchr(soname, '/');
	if (basename) {
		++basename;
	}
	else {
		basename = soname;
	}
	strcat(strcpy(vecname, "__V_"), basename);
    {
	char *p;
	for (p = vecname; 0!=*p; ++p) {
		if (!isalnum(*p)) {
			*p = '_';
		}
	}
    }
	return vecname;
}

regex_t *
make_regex(char const *const *tokens, int n_tokens)
{
	regex_t *retval, *rp;
	rp = retval = (regex_t *)zalloc((1+n_tokens) * sizeof(regex_t));

	for (; *tokens; ++rp, ++tokens) {
		int const err = regcomp(rp, *tokens, REG_EXTENDED|REG_NOSUB);
		if (0!=err) {
			char errbuf[400];
			size_t esize = regerror(err, rp, errbuf, sizeof(errbuf));
			fprintf(stderr, "*** %s\n", errbuf); (void)esize;
			fprintf(stderr, "regexp was `%s'\n", *tokens);
			exit(1);
		}
	}
	memset(rp, 0, sizeof(*rp));
	return retval;
}

char const *const *
get_tokens(char *const arg, size_t const size, int *n_tokens)
{
	char const *const argend = size + arg;
	char *p;
	int j=2;  /* 1 for terminator, 1 for no'\n' at end */
	for (p=arg; p<argend; ++p) if (isspace(*p)) {
		++j;
	}
  {
	char const **const tokens = (char const **)zalloc(j*sizeof(*tokens));
	char const **pp = tokens;
	for (p=arg; p<argend; ) {
		while (isspace(*p)) {
			++p;
		}
		if ('#'==*p) {  /* skip over comment */
			while (p < argend && '\n'!=*p) {
			 	++p;
			}
			continue;
		}
		*pp++ = p;  /* start of token */
		while (p < argend && !isspace(*p)) {
			if ('#'==*p) { /* skip comment */
				*p = 0;  /* but terminate preceding token */
				while (p < argend && '\n'!=*p) {
			 		++p;
				}
			}
			++p;
		}
		if (0!=*p) {  /* allow read-only single string */
			*p = 0;
		}
		++p;
	}
	*pp = 0;  /* terminator */
	*n_tokens = pp - tokens;
	return tokens;
  }
}

void
free_symbol_patterns(regex_t *patterns)
{
	regex_t *rp = patterns;
	for (; rp->buffer; ++rp) {
		regfree((void *)rp);
	}
	free(rp->fastmap); rp->fastmap = 0;
	regfree((void *)rp);
	free(patterns);
}

regex_t *
get_symbol_patterns(char *arg, int no_regex)
{
	size_t len = strlen(arg);
	if ('@' == arg[0]) {
		arg = getfile(1+arg, &len, 1);
	}
  {
	int n_tokens;
	char const *const *const tokens = get_tokens(arg, len, &n_tokens);
	if (0==no_regex) {
		regex_t *const rp = make_regex(tokens, n_tokens);
		/* hold tokens for free(); rp[n_tokens].buffer==0 is sentinel for end */
		rp[n_tokens].fastmap = (void *)tokens;
		return rp;
	}
	else {
		return (regex_t *)tokens;
	}
  }
}

int
matches(
	char const *const name,
	regex_t const *pat  /* array of patterns */
)
{
	for (; 0!=pat->buffer; ++pat) {
		if (0==regexec(pat, name, 0, 0, 0)) {
			return 1;
		}
	}
	return 0;
}

void
veclib_retrofit(
	int const fdo,
	Elf32_Ehdr *const ehdr,
	char *const ante_hash,
	size_t const sz_ante_hash,
	unsigned *const new_htab,
	Elf32_Sym *const post_sym,
	int n_post_sym
)
{
	Elf32_Phdr const *const pt_dynamic = elf_find_phdr_type(PT_DYNAMIC, ehdr);
	Elf32_Dyn *const dynamic = 
		(Elf32_Dyn *)(pt_dynamic->p_offset + (char *)ehdr);
	unsigned *const htab = (unsigned *)elf_find_dynamic(DT_HASH, dynamic, ehdr);
	Elf32_Sym *const dynsym = (Elf32_Sym *)elf_find_dynamic(DT_SYMTAB, dynamic, ehdr);
	char const *const dynstr = (char const *)elf_find_dynamic(DT_STRTAB, dynamic, ehdr);
	size_t const dt_strsz = elf_find_section(dynstr, ehdr)->sh_size;
		/* elf_get_dynamic(DT_STRSZ, dynamic) omits some */

	unsigned const dt_versym = elf_get_dynamic(DT_VERSYM, dynamic);
	unsigned short *const versec = (unsigned short *)(
		((~0!=dt_versym)
			? elf_find_offset(dt_versym, ehdr)
			: 0 ) );
	/* write new .so to stdout */
	Elf32_Shdr *shdr = (Elf32_Shdr *)(ehdr->e_shoff + (char *)ehdr);
	off_t offset = sizeof(*ehdr)
	             + ehdr->e_phnum * sizeof(Elf32_Phdr)
	             + ehdr->e_shnum * sizeof(Elf32_Shdr);
	char *addstrs = 0;
	size_t addchars = 0;
	int j;
	
	uwrite(fdo, ehdr, offset);
  for (j=0; j<ehdr->e_shnum; ++shdr, ++j) {
		char const *const sptr = shdr->sh_offset + (char const *)ehdr;
	if (sptr==(char const *)htab) {
		if ((shdr[1].sh_offset + (char const *)ehdr)!=(char const *)dynsym
		||  (shdr[2].sh_offset + (char const *)ehdr)!=(char const *)dynstr
		||  (~0U!=dt_versym
		    && (char *)versec!=(shdr[3].sh_offset + (char *)ehdr) ) ) {
			fprintf(stderr,"not in order  SHT_HASH, SHT_DYNSYM, SHT_STRTAB, SHT_GNU_versym\n");
			exit(1);
		}
		offset  = ulseek(fdo,  shdr->sh_offset, SEEK_SET);
		offset += uwrite(fdo, ante_hash, sz_ante_hash);
		shdr->sh_addr = offset + (shdr->sh_addr - shdr->sh_offset);
		shdr->sh_offset = offset;
		elf_set_dynamic(DT_HASH, shdr->sh_addr, dynamic);
		shdr->sh_size = sizeof(unsigned)*(2 + new_htab[0] + new_htab[1]);
		offset += uwrite(fdo, new_htab, shdr->sh_size);
	}
	else if (sptr==(char const *)dynsym) {
		size_t const addsyms = sizeof(Elf32_Sym) * n_post_sym;
		off_t const oldoff = shdr->sh_offset;
		shdr->sh_addr = offset + (shdr->sh_addr - shdr->sh_offset);
		shdr->sh_offset = offset;
		elf_set_dynamic(DT_SYMTAB, shdr->sh_addr, dynamic);
		offset += uwrite(fdo, oldoff + (char *)ehdr, shdr->sh_size);
		{
			Elf32_Sym *symp;
			char *cp;
			int j;
			for (symp=post_sym, j=n_post_sym; --j>=0; ++symp) {
				addchars += 1+ strlen((char *)symp->st_name);
			}
			cp = addstrs = (char *)malloc(addchars);
			addchars = 0;
			for (symp=post_sym, j=n_post_sym; --j>=0; ++symp) {
				char const *const t = (char const *)symp->st_name;
				size_t const len = 1+ strlen(t);
				symp->st_name = addchars + dt_strsz;
				addchars += len;
				memcpy(cp, t, len);
				cp += len;
			}
		}
		offset += uwrite(fdo, post_sym, addsyms);
		shdr->sh_size += addsyms;
	}
	else if (sptr==dynstr) {
		off_t const oldoff = shdr->sh_offset;
		shdr->sh_addr = offset + (shdr->sh_addr - shdr->sh_offset);
		shdr->sh_offset = offset;
		elf_set_dynamic(DT_STRTAB, shdr->sh_addr, dynamic);
		offset += uwrite(fdo, oldoff + (char *)ehdr, shdr->sh_size);
		offset += uwrite(fdo, addstrs, addchars);
		elf_set_dynamic(DT_STRSZ, shdr->sh_size += addchars, dynamic);
	}
	else if (SHT_GNU_versym==shdr->sh_type) {
		unsigned short const vecver = 1;
		int k;
		offset  = ulseek(fdo, shdr->sh_offset -
			n_post_sym * sizeof(vecver), SEEK_SET);
		elf_set_dynamic(DT_VERSYM, dt_versym -
			n_post_sym * sizeof(vecver), dynamic);
		offset += uwrite(fdo, shdr->sh_offset +(char *)ehdr, shdr->sh_size);
		for (k=n_post_sym; --k>=0; ) {
			offset += uwrite(fdo, &vecver, sizeof(vecver));
			shdr->sh_offset -= sizeof(vecver);
			shdr->sh_addr   -= sizeof(vecver);
			shdr->sh_size   += sizeof(vecver);
		}
	}
	else {
		offset  = ulseek(fdo, shdr->sh_offset, SEEK_SET);
		offset += uwrite(fdo, shdr->sh_offset +(char *)ehdr, shdr->sh_size);
	}
  }
	free(addstrs);
	ulseek(fdo, ehdr->e_shoff,  SEEK_SET);
	uwrite(fdo, ehdr->e_shoff + (char *)ehdr, ehdr->e_shnum * sizeof(Elf32_Shdr));
}

